$ifndef _AMICUS18_GPS_
$define _AMICUS18_GPS_
'
' Proton GPS Receiver routines
' Receive GPS data and parse it into date, time, and position
'
' Written by Les Johnson. 3rd January 2016
'
' Setup the software serial port to receive data from the GPS receiver
'
    Declare Serial_Baud = 9600                      ' GPS receiver baud rate
    Declare Rsin_Timeout = 1000                     ' Set the Rsin timeout to 1 second (1000ms)
    Declare Rsin_Pin = PORTA.0                      ' From GPS receiver TX line
    Declare Rsin_Mode = True                        ' Non-Inverted serial data
'
' Variables used in the GPS_StrToFloat subroutine
'
    Dim Atf_fPow10 As Float System
    Dim Atf_fResult As Float System
    Dim Atf_tSign As Bit
    Dim PP_AARG As Float System
    Dim GPS_wFSR0 As FSR0L.Word
'
' Value extraction variables
'
    Dim GPS_bLoop As Atf_fPow10.Byte0               ' General purpose loop variable
    Dim GPS_bExtractItem As Atf_fPow10.Byte1
    Dim GPS_bCommaCount As Atf_fPow10.Byte2
    Dim GPS_bCharpos As Atf_fPow10.Byte3
    Dim GPS_bChar As Byte System
'
' GPS variables
'
    Dim GPS_Day As Byte                             ' Holds the Day value
    Dim GPS_Month As Byte                           ' Holds the Month value
    Dim GPS_Year As Byte                            ' Holds the Year value
    Dim GPS_Hours As Byte                           ' Holds the Hours value
    Dim GPS_Minutes As Byte                         ' Holds the Minutes value
    Dim GPS_Seconds As Byte                         ' Holds the Seconds value
    Dim GPS_Latitude As Float                       ' Holds the Latitude value
    Dim GPS_LatBearing As Byte                      ' Holds the Latitude Bearing character (N or S)
    Dim GPS_Longitude As Float                      ' Holds the Longitude value
    Dim GPS_LonBearing As Byte                      ' Holds the Longitude Bearing character (E or W)
    Dim GPS_FixType As Byte System                  ' Hold the GPS fix type (0, 1, 2 or 3)
    Dim GPS_SatsInView As Byte                      ' Holds the amount of satellites in view

    Dim GPS_bDataString As String * 20 Heap         ' Temporary string
    Dim GPS_bRxBuffer[84] As Byte Heap              ' Holds the received characters from the GPS module

'------------------------------------------------------------------------------

    GoTo _GPS_Main_                                 ' Jump over the GPS subroutines

'------------------------------------------------------
' Convert an ASCII representation of a Floating Point value into a Floating Point variable
' Input     : GPS_bDataString holds the ASCII representation of the floating point value
' Output    : Atf_fResult holds the converted floating point value
' Notes     : Recognises -, +, and . as the beginning of the ASCII floating point value
'
GPS_StrToFloat:
    Atf_fPow10 = 1
    Atf_fResult = 0

    GPS_wFSR0 = AddressOf(GPS_bDataString)          ' Point FSR0L/H to the start of the string
    GPS_bChar = POSTINC0                            ' Get a character from the String
    Atf_tSign = 0
    If GPS_bChar = "-" Then                         ' Have we found a "-" character?
        GPS_bChar = POSTINC0                        ' Yes. So Discard the "-" character
        Atf_tSign = 1                               ' Indicate that the value is negative
    Else If GPS_bChar = "+" Then                    ' Have we found a "+" character?
        GPS_bChar = POSTINC0                        ' Yes. So Discard the "+" character
    EndIf
    While                                           ' Scan the digits before the decimal point (if included)
        Select GPS_bChar                            ' Exit the loop if non-numeric characters found
            Case <"0", >"9"
                GoTo StrToFlt_Skp1
        EndSelect
        Atf_fResult = Atf_fResult * 10              ' \
        PP_AARG = Atf_fResult + GPS_bChar           ' | Calculate the whole part of the floating point value
        Atf_fResult = PP_AARG - "0"                 ' /
        GPS_bChar = POSTINC0                        ' Get another character from the string
    Wend
StrToFlt_Skp1:
    If GPS_bChar == "." Then                        ' Have we found a "." character?
        GPS_bChar = POSTINC0                        ' Yes. So Discard the "." character
        While                                       ' Scan the digits after the decimal point (if included)
            Select GPS_bChar                        ' Exit the loop if non-numeric characters found
                Case <"0", >"9"
                    GoTo StrToFlt_Skp2
            EndSelect
            Atf_fPow10 = Atf_fPow10 * 10            ' \
            PP_AARG = GPS_bChar - "0"               '  | Calculate the fractional part of the floating point value
            PP_AARG = PP_AARG / Atf_fPow10          '  |
            Atf_fResult = Atf_fResult + PP_AARG     ' /
            GPS_bChar = POSTINC0                    ' Get another character from the string
        Wend
    EndIf
StrToFlt_Skp2:
    If Atf_tSign = 1 Then Atf_fResult = -Atf_fResult ' Convert to negative if required
    Return

'------------------------------------------------------------------------------
' Extract a value based on the comma that precedes it
' Locate and extract the value following the comma number placed in GPS_bExtractItem
' Input     : GPS_bExtractItem holds the comma value to find the value after
'           : Array GPS_bRxBuffer holds the GPS text string
' Output    : The extracted text is returned in string GPS_bDataString
' Notes     : None
'
$define GPS_ExtractValue(pCommaPosition) '
    WREG = pCommaPosition '
    GoSub __GPS_ExtractValue

__GPS_ExtractValue:
    GPS_bExtractItem = WREG
    Clear GPS_bDataString                               ' Clear the resulting string before we start
    GPS_bCommaCount = 0                                 ' Reset the comma counting variable
    GPS_bCharpos = 0                                    ' Start at the beginning of the array
    Repeat                                              ' Form a loop to scan the array
        GPS_bChar = GPS_bRxBuffer[GPS_bCharpos]         ' Scan the buffer array to parse
        If GPS_bChar = "," Then Inc GPS_bCommaCount     ' Increment GPS_bCommaCount if a comma is found
        If GPS_bCommaCount = GPS_bExtractItem Then      ' Have we found the correct comma?
            GPS_bLoop = 0                               ' Yes.  So....
            Repeat                                      ' Create a loop
                Inc GPS_bCharpos                        ' Skip over the comma and keep scanning the array
                GPS_bChar = GPS_bRxBuffer[GPS_bCharpos] ' Extract the pieces of the value into GPS_bChar
                Select GPS_bChar                        ' Have we found a terminator character?
                    Case 0, 13, ","                     ' Yes. So...
                        Return                          ' Exit the subroutine
                EndSelect
                GPS_bDataString[GPS_bLoop] = GPS_bChar  ' Fill GPS_bDataString with the value
                Inc GPS_bLoop                           ' Point to the next data piece
            Until GPS_bCharpos >= SizeOf(GPS_bRxBuffer) ' Keep looping until a terminator is found, or the array runs out
            Return
        EndIf
        Inc GPS_bCharpos                                ' Move to the next byte in the array
    Until GPS_bCharpos >= SizeOf(GPS_bRxBuffer)         ' Loop until all the elements of the array are scanned
    Return

'------------------------------------------------------------------------------
' Receive the line of NMEA text from the GPS module
' Input     : Received serial data from the GPS module
' Output    : GPS_bRxBuffer holds the line of NMEA text
' Notes     : None
'
_GPS_GatherData:
    GPS_wFSR0 = AddressOf(GPS_bRxBuffer)                ' Point FSR0L/H to the start of the buffer
    GPS_bCharpos = SizeOf(GPS_bRxBuffer)
    Repeat                                              ' Create a loop to receive the serial string
        Dec GPS_bCharpos
        RSIn {__GPS_Timeout},GPS_bChar                  ' Receive a character serially from the GPS module
        If GPS_bChar = 13 Then Break                    ' Exit the loop when the string reaches the end
        POSTINC0 = GPS_bChar                            ' Load array GPS_bRxBuffer
    Until GPS_bCharpos = 0                              ' Repeat the loop until the buffer runs out
    Return

'------------------------------------------------------------------------------
' Convert the byte elements of string GPS_bDataString to integer values
' Input     : GPS_bDataString holds the ASCII value
' Output    : GPS_bDataString holds the integer values of the characters
' Notes     : None
'
_GPS_ElementsToInt:
    GPS_wFSR0 = VarPtr GPS_bDataString                  ' Point FSR0L/H to the start of the string
    While                                               ' Create a loop
        GPS_bChar = INDF0                               ' Get a byte from the string
        If GPS_bChar = 0 Then Return                    ' Exit the subroutine when the string reaches the end
        POSTINC0 = GPS_bChar - 48                       ' Load GPS_bDataString with the integer equivalent of the character
    Wend
    Return

'------------------------------------------------------------------------------
' Parse GPS data
' Extract the location, time and date information from the GPS module
' Input     : None
' Output    : GPS_Day Holds the Day value
'           : GPS_Month Holds the Month value
'           : GPS_Year Holds the Year value
'           : GPS_Hours Holds the Hours value
'           : GPS_Minutes Holds the Minutes value
'           : GPS_Seconds Holds the Seconds value
'           : GPS_Latitude Holds the floating point Latitude value
'           : GPS_LatBearing Holds the Latitude Bearing character (N or S)
'           : GPS_Longitude Holds the floating point Longitude value
'           : GPS_LonBearing Holds the Longitude Bearing character (E or W)
'           : GPS_FixType Hold the GPS fix type (0, 1, 2 or 3)
'           : GPS_SatsInView Holds the amount of satellites in view
' Notes     : None
'
$define GPS_Parse() GoSub __GPS_Parse

__GPS_Parse:
    RSIn {__GPS_Timeout},Wait("$GPGGA")                 ' Wait for the $GPGGA string (with a timeout)
    GoSub _GPS_GatherData                               ' Receive the rest of the NMEA text into GPS_bRxBuffer
'
' Place the time and position values into their respective variables
'
    GPS_ExtractValue(1)                                 ' Get time ASCII value into GPS_bDataString
    GoSub _GPS_ElementsToInt                            ' Convert the lements into integer values
    GPS_Hours = GPS_bDataString#1 + (GPS_bDataString#0 * 10)    ' Extract the Hours value
    GPS_Minutes = GPS_bDataString#3 + (GPS_bDataString#2 * 10)  ' Extract the Minutes value
    GPS_Seconds = GPS_bDataString#5 + (GPS_bDataString#4 * 10)  ' Extract the Seconds value

    GPS_ExtractValue(2)                                 ' Get the latitude ASCII value into GPS_bDataString
    GoSub GPS_StrToFloat                                ' Convert the latitude ASCII value into a floating point variable
    GPS_Latitude = Atf_fResult

    GPS_ExtractValue(3)                                 ' Get the latitude bearing ASCII value into GPS_bDataString
    GPS_LatBearing = GPS_bDataString#0

    GPS_ExtractValue(4)                                 ' Get the longitude ASCII value into GPS_bDataString
    GoSub GPS_StrToFloat
    GPS_Longitude = Atf_fResult                         ' Convert the longitude ASCII value into a floating point variable

    GPS_ExtractValue(5)                                 ' Get the longitude bearing ASCII value into GPS_bDataString
    GPS_LonBearing = GPS_bDataString#0

    GPS_ExtractValue(6)                                 ' Get the fix type ASCII value into GPS_bDataString
    GPS_FixType = GPS_bDataString#0 - 48                ' Convert it into an integer

    GPS_ExtractValue(7)                                 ' Get the amount of satellites ASCII value into GPS_bDataString
    GPS_SatsInView = GPS_bDataString#0 - 48             ' Convert it into an integer
'
' Find the date of the GPS fix
'
    RSIn {__GPS_Timeout},Wait("$GPRMC")                 ' Wait for the $GPRMC string (with a timeout)
    GoSub _GPS_GatherData                               ' Receive the rest of the NMEA text into GPS_bRxBuffer
'
' Place the date information into its respective variables
'
    GPS_ExtractValue(9)                                ' Extract the date information into the GPS_bDataString array
    GoSub _GPS_ElementsToInt                           ' Convert the lements into integer values
    GPS_Day = GPS_bDataString#1 + (GPS_bDataString#0 * 10)     ' Extract the Day value
    GPS_Month = GPS_bDataString#3 + (GPS_bDataString#2 * 10)   ' Extract the Month value
    GPS_Year = GPS_bDataString#5 + (GPS_bDataString#4 * 10)    ' Extract the Year value
    Return
'
' Come here if a serial timeout occurs
'
__GPS_Timeout:
    GPS_FixType = 0                                     ' Indicate no satellites in view
    Return

'------------------------------------------------------------------------------
_GPS_Main_:

$endif
